unit mystuff;

{$mode objfpc}{$H+}{$R+}{$I+}

//--------------------------------------------------------------------------
//
//  str_strings module
//
//  This module contains my standardized routines which includes:
//
//  Prefix:
//      fil                             file access routines
//      str                             string access routines
//      stt                             standard types
//      <none>                          Various misc routines
//
// *** File stuff ***
//
// bool := fil_delete(fname);           Delete File. Return value always false?
// bool := fil_direxists(dirname);      Does directory exist?
// bool := fil_exists(fname);           Does File exist?
// s := fil_getprogdir;                 Get directory of program file
// s := fil_qfname(fname);              Fully qualifies a filename (like FEXPAND)
// fil_redirect;                        Allows redirection of input and output from DOS.
// err := fil_rename(old,new);          Rename a file.
// l := fil_size(fname, err);           Return size of file in bytes.
// fil_splitname(path, dir, name, ext); Like FSPLIT, but uses regular strings.
// s := fil_tempfile(path);             returns name of a temporary file in
//                                      the specified directory.

//                                      send an HTML formatted email
// b:=mai_SendHTML(tolist,cclist,bcclist,from,subject,msgLineCount,msg);

// *** Misc stuff ***

// b := isbetween(x,i1,i2)              is  i1 <= x <= i2 ?
// memdump(v, sizeof(v));               Formatted dump of a variable
// memdump_tsl(v, sizeof(v), tsl);      Formatted dump of a variable to tStringList
// noop;                                no-op - does nothing

// *** StackDump notes:
// There are 2 'types' of stackdumps and 2 ways the output can be formatted
// for a total of 12 calls.

// StackDump_tsl is the primary call. It will dump the stack markers to a TStringList.
// StackDump_tsl(e) also dumps stack markers but also include info about an exception.
// stackDump outputs the tsl from stackDump_tsl to stdlist.
// stackDump(e) outputs the tsl from stackDump_tsl(e).

// stackDump;
// stackDump(e);                        Do a stack dump, pass exception.
// stackDump_tsl(tsl);
// stackDump_tsl(e,tsl);                stack dump to tStringList

// *** packed array of byte manip routines. I'm using PABs in sjsconfig
// *** as fixed fields containing characters
// s := pab2s(pab);                     convert packed array of byte to string;
// s2pab(pab, s);                       convert string to pab (trailing spaces)
// s2pab0(pab, s);                      convert string to pab (trailing 0's)

// *** String stuff ***

// Aos type is an Array of strings

// str_AoSdispose(AoS);                 dispose of all strings in AoS
// str_AoSdump(AoS);                    display all strings in AoS
// str_AoSinit(AoS);                    Initialize all ptr's to NIL
// s:= str_AoSget(AoS[n]);              Retrieve an AoS string;
// str_AoSput(AoS[n],s);                Create an AoS string;

// b := str_contains(s1,s2);            returns true if s2 in s1 (case insensitive)
// n := str_count(s, substr);           Count occurrences of <substr> in s
// s := str_date(d);                    returns mm/dd/yyyy hh:mm:ss.mmmm
// s := str_datetime(d);                returns mm/dd/yyyy hh:mm:ss
// s := str_date2hhmm(d);               returns hh:mm:ss
// s := str_date2mdy(d);                returns mm/dd/yyyy
// s := str_date2wwwdd(d);              returns www dd such as MON 01
// s := str_deb(s);                     Remove leading and trailing blanks.
// s := str_delete(s,i,n)               Delete characters in s starting at
//                                      s[i] for <n> characters
// str_delimit(s,sep,delim,n,AoS);      Parse s into AoS using separator and delimiter
// str_delimit_TSL                      Like str_delimit, but uses TStringList rather than AoS
// s := str_downshift(s);               Downshift string
// s := str_fmtint(i,n);                Cvrt Integer to String. n = field size.
//                                      0 = free format.
// s := str_fmtlong(l, n);              Cvrt Long to string. n = field size.
//                                      0 = free format.
// n := str_getbits(i,p,n);             Given i, return bits @ <p> for <n> bits.
// i := str_hex2int(s);                 Given hex string, return integer value.
// i := str_inet_aton(s);               NOT IMPLEMENTED
// s := str_inet_ntoa(i);               Given integer IP addr, return dotted form.
// s := str_int2commas(i);              Given i, return a string nnn,nnn,nnn
// s := str_int2hex(i,n);               Given i, convert to hex string of size <n>.
// s := str_int2str(i);                 Convert integer <i> to a string.
// b := str_isnumeric(s);               Is <s> only digits?
// s := str_justl(s,n);                 Left just string to fit in n characters
// s := str_justr(s,n);                 Right Just string to fit in n characters
// s := str_len(s);                     Just like Length(s);
// s := str_lz(s);                      Convert leading spaces to leading zeros
// p := str_nextid(s,p,start,elem,type,id); Parse next identifier.
// s := str_pac2str(s);                 Convert packed array of char to string
// s := str_padl(s,n);                  if len(s)<n, pad leading spaces
// s := str_padr(s,n);                  if len(s)<n, pad trailing spaces
// p := str_pos(s,substr);              returns position of <substr> in <s>. 0 is
//                                      returned if not found.
// s := str_properCase(s, delims);      string converted to lowercase. 1st char uppercase.
// n := str_putbits(i,p,n,v);           write <v> into <i> @ bit <p> for <n> bits.
//                                      returns -1 if fails.
// p := str_quotedstr(s,p,qs,err);      parses quoted string (including
//                                      imbedded quotes).
// s := str_range(s, start, stop);      copy <s[start]>..<s[stop]>
// s := str_real2commas(r,n);           Given r, return a string nnn,nnn,nnn.nn
// s := str_repeat(c,n);                Create str of <n> <c>'s
// s := str_replace(s,f,t);             Replace <f> w/ <t> in <s>
// s := str_rot13(s);                   encode string in ROT13
// s := str_rot47(s);                   Like ROT13, but encodes numeric and special chars
// i := str_scan(s,i,inflag,cset);      Scans a string. See procedure for explanation
// i := str_setlen(s,len);              Forces length of a string
//                                      (returns old length).
//                                      If new len is > old len, new area is not
//                                      initialized.
// i := str_str2int(s);                 converts <s> to integer <i>. If <s> is not an
//                                      integer, 0 is returned.
// r := str_str2real(s);                converts <s> to real.
// str_str2pac(s, pac);                 Moves string into packed array of char.
// str_str2pac0(s, pac);                Like str2pac, but adds #0 terminator
// s := str_striphtml(s);               Given a string, this strips html which is
//                                      simply all <...>.
// s := str_substr(s, start, len);      copy <s[start]> for <len> characters
//                                      (like copy)
// s := str_trim(s);                    same as str_deb
// s := str_triml(s);                   Trim All spaces to the left
// s := str_trimr(s);                   Trim All spaces to the right
// s := str_truncl(s,n);                if len(s) > n, set to n, truncating from
//                                      left.
// s := str_truncr(s,n);                if len(s) > n, set to n, truncating from
//                                      right.
// s := str_upshift(s);                 Upshift String
// n := str_unstring(s,strArr,[sep,     Simpler str_delim. unstring <s> into
//          [delim]]);                  array of string; n is #str parsed.
// i := str_wrap(s, p, len);            return index where word wrap should occur

// 5.18 May 2022: BigDan
//      Enh: added stt_bits type
// 5.17 Sep 2019: BigDan
//      Enh: added str_properCase.
// 5.16 Jul 2019: BigDan
//      Enh: added noop.
// 5.15 Dec 2018: BigDan
//      Enh: added s2pab0.
// 5.14 Nov 2018: BigDan
//      Enh: added stackdump and stackDump_tsl (without exception parm).
// 5.13 Jan 2017: BigDan
//      Enh: added str_str2pac0 to convert string to PAC w/ trailing #0.
//      Enh: str_trimR, and by extension str_deb, will now properly convert
//          PAC w/ trailing #0 into a string w/o trailing #0. This allows
//          you to use str_trimR or str_deb to directly convert PAC to
//          string rather than using str_pac2str which is a proc not a func.
//      Enh: added pab2s and s2pab to manipulate packed arrays of byte. FPCs
//          handling of PAC can be problematic.
// 5.12 Oct 2017: BigDan
//      Enh: added str_pac2str to convert packed array of char to string.
// 5.11 Jul 2017: BigDan
//      Fix: mai_SendHTML is working again, but only on unix.
//      Enh: fil_tempfile now uses the fpc internal getTempFileName.
//      Enh: added memdump_tsl.
//      Enh: added stackdump_tsl.
//      Enh: added str_datetime
// 5.10 Mar 2017: BigDan
//      Enh: added str_contains.
//      Enh: added memDump;
//      Enh: added stackDump;
// 5.9 Sep 2016: BigDan
//      Fix: changing str_range's last parm to be a default of 0 bit me in the
//      butt. str_delimit will call str_range with a count of 0 and expect NOTHING
//      to be copied (a null string is returned). Instead, the new version would
//      return the whole string.
//      NOW, if the last parm of str_range is omitted, it is set to -1 which should
//      really be invalid. This won't impact anything using the omitted parameter,
//      and anyone that really wants 0 bytes xferred will get 0 bytes.
// 5.8 May 2016: BigDan
//      Enh: Added isbetween function
// 5.7 Mar 2016: BigDan
//      Fix: In str_delimit and str_delimit_TSL, the function delimit would
//      abort if there was a null string between seperators such as a,,b.
//      Enh: str_range's <stop> parameter is now optional. If omitted, it is
//      the same as writing str_range(s, n, str_len(s)), or it returns the entire
//      substring after n.
// 5.6 Feb 2016: BigDan
//      Enh: Add str_wrap to handle word wrapping.
//      Fix: lazarus 1.6 gets range error in function length due to definition of
//      stt_int4. This was set as -n .. n. Now I just use int32 and I also changed
//      all other stt_int<n> defs to match.
//      Enh: added:
//          str_date, str_date2hhmm, str_date2mdy(d), str_date2wwwdd(d)
// 5.5 Dec 2015: BigDan
//      Fix: the type stt_int8 was equated to longint. Turns out that was a 32 bit
//      integer, NOT 64 bit. It is now equated to int64 which is 64 bits.
// 5.4 Jul 2015: BigDan
//      Enh: added str_unstring. str_delim is always hard to remember how
//      to use. str_unstring is a simpler interface to str_delim. To use
//      create strArr: array [1..n] of string. Then simply call with
//      str_unstring('a,b,c',strArr). strArr[1] contains 'a', etc.
// 5.3 Feb 2015: BigDan
//      Fix: str_fmtint and str_hextoint had stt_int4 in forward, but stt_int8
//      in actual procedure which caused compile error.
//      Commented out mai routine. It has problems and there is no tester.
//      If I find I need it, I can debug it and add it back.
//      Enh: implemented stt_int<n>b and stt_card<n>b types.
//      Enh: added str_delimit_TSL which uses TStringList rather than AoS
//      Enh: added str_rot13 and str_rot47
// 5.2 Sep 2013: BigDan
//      Fix: when compiling on RPI, stt_int4 was not using 4 byte integers.
//      Now I specify the exact integer range to make sure I get the right
//      number of bytes.
// 5.1 Aug 2013: BigDan
//      Enh: ported the AoS string stuff
//      Enh: ported mai_sendHTML procedure
// 5.0 Mar 2012: BigDan
//      Enh: Port to FreePascal.
//      Converted any stt_int4 passed by value to stt_int8.
//      removed str_fmtlong as it is no longer needed.
//      dropped the AoS stuff since fp implements 255 byte string.
//      added str_trim, str_str2pac.
// 4.3 Jun 2006: BigDan
//      Enh: added str_striphtml
// 4.2 Sep 2005: BigDan
//      Enh: added some new routines to support the str_eiip routines I'm
//      working on:
//         str_int2hex    convert integer to string in hex format
//         str_hex2int    convert hex in string to an integer
//         str_inet_aton  convert dotted IP addr (x.x.x.x) to integer.
//         str_inet_ntoa  convert IP addr in integer fmt to dotted fmt.
//      str_get/putbits isn't exactly string manipulation, but I don't have
//           another module to put these into.
//           str_getbits    retrieve a bit field from a 32 bit cardinal.
//           str_putbits    write a bitfield into a 32 bit cardinal.
// 4.1 Aug 2004: BigDan
//      Enh: added routines to handle Arrays of String Pointers (AoS).
// 4.0 Jun 6, 2004: BigDan
//      Enh: Port to Linux and GPC
// 3.0 Dec 11, 2003: BigDan
//      enh: Updated to use in Delphi7
// 2.0 Aug 23, 1995: BigDan
//      enh: Compiled in TP 7.0 which allowed me to use the delay procedure
//          in sy_delay.
// 1.0 Aug 26, 1994: BigDan
//      Original Version:
//          fi_getprogdir,  str_downshift, str_justl, str_justr, str_padl
//          str_padr, str_repeat, str_triml, str_trimr, st_trunl, str_truncr,
//          str_upshift, sy_enabletrap, sykeybuf, sy_keypressed,
//          sy_system.
//
//--------------------------------------------------------------------------

interface

uses
    classes,                                                // for TStringList
    sysutils;

type
    str_charset                         = set of char;

    // stt_bits<n> where <n> is number of bits. To be used with bitpacked record
    stt_bits1                           = 0..1;
    stt_bits2                           = 0..3;
    stt_bits3                           = 0..7;
    stt_bits4                           = 0..15;
    stt_bits5                           = 0..31;
    stt_bits6                           = 0..63;
    stt_bits7                           = 0..127;
    stt_bits8                           = 0..255;
    stt_bits9                           = 0..511;
    stt_bits10                          = 0..1023;
    stt_bits11                          = 0..2047;
    stt_bits12                          = 0..4095;
    stt_bits13                          = 0..8191;
    stt_bits14                          = 0..16383;
    stt_bits15                          = 0..32767;
    stt_bits16                          = 0..65535;

{   Bitpacked record notes:
    To use bitpacked fields, you need the rec of bit fields and it needs to
    be redefined by a cardinal (unsigned int). Set the bits directly, but
    use the cardinal for passing to other procedures. For example:
        rec                             : bitpacked record
            case integer of
                0: (
                    a                   : stt_bits3;
                    b                   : stt_bits3;
                    c                   : stt_bits2;
                    );
                1: (
                    x                   : stt_card8b;
                    );
            end;
        If you do a memdump(rec) and compare it to inttohex(rec.x) you
        will find rec.x contains what you expect.
}

    // stt_card<n> where <n> is # of bytes
    stt_card1                           = 0..255;           // 8 bit cardinal
    stt_card2                           = 0..65535;         // 16 bit cardinal
    stt_card4                           = 0..4294967295;    // 32 bit cardinal
    stt_card8                           = 0..18446744073709551615;

    // stt_card<n>b where <n> is # of bits
    stt_card8b                          = stt_card1;
    stt_card16b                         = stt_card2;
    stt_card32b                         = stt_card4;
    stt_card64b                         = stt_card8;

    // stt_int<n> where <n> is # of bytes
    stt_int1                            = int8;             // 8 bit (byte) integer
    stt_int2                            = int16;            // 16 bit integer
    stt_int4                            = int32;            // 32 bit integer
    stt_int8                            = int64;            // 64bit integer

    // stt_int<n>b where <n> is # of bits
    stt_int8b                           = stt_int1;
    stt_int16b                          = stt_int2;
    stt_int32b                          = stt_int4;
    stt_int64b                          = stt_int8;

    stt_pstr                            = ^string;
    stt_aos                             = array of stt_pstr; //array of pointers to strings

    EStdUnit                            = class(Exception);
    EMaiRuncommand                      = class(EStdUnit);
    EMaiWindows                         = class(EStdUnit);

function fil_delete(
        fname                           : string
    )                                   : boolean;

function fil_direxists(
        dirname                         : string
    )                                   : boolean;

function fil_exists(
        fname                           : string
    )                                   : boolean;

function fil_getprogdir                 : string;

function fil_qfname(
        s                               : string
    )                                   : string;

procedure fil_redirect;

function fil_rename(
        old                             : string;
        new                             : string
    )                                   : boolean;

function fil_size(
        fname                           : string;
    var err                             : stt_int4
    )                                   : stt_int8;

procedure fil_splitname(
        path                            : string;
    var dir                             : string;
    var name                            : string;
    var ext                             : string
    );

function fil_tempfile(
        path                            : string
    )                                   : string;


function mai_SendHTML(
        tolist                          : string;
        cclist                          : string;
        bcclist                         : string;
        from                            : string;
        subject                         : string;
        msgLineCount                    : stt_int4;
        msg                             : stt_aos
    )                                   : boolean;

function isbetween(
        x                               : stt_int8;
        loBound                         : stt_int8;
        hiBound                         : stt_int8
    )                                   : boolean;

procedure memDump(
        v                               : pointer;
        size                            : stt_int32b
    );

procedure memDump_tsl(
        v                               : pointer;
        size                            : stt_int32b;
    var tsl                             : TStringList
    );

procedure noop;

function pab2s(
        pab                             : array of byte
    )                                   : string;

procedure s2pab(
    var pab                             : array of byte;
        s                               : string
    );

procedure s2pab0(
    var pab                             : array of byte;
        s                               : string
    );

procedure stackDump;

procedure stackDump(
        e                               : Exception
    );

procedure stackdump_tsl(
    var tsl                             : TStringList
    );

procedure stackDump_tsl(
        e                               : Exception;
    var tsl                             : TStringList
    );

procedure str_AoSdispose(
    var AoS                             : stt_AoS
    );

procedure str_AoSdump(
    var AoS                             : stt_AoS
    );

procedure str_AoSinit(
    var AoS                             : stt_AoS
    );

function str_AoSget(
    var AoSn                            : stt_pstr
    )                                   : string;

procedure str_AoSput(
    var AoSn                            : stt_pstr;
        s                               : string
    );

function str_contains(
        s1                              : string;
        s2                              : string
    )                                   : boolean;

function str_count(
        s                               : string;
        substr                          : string
    )                                   : stt_int4;

function str_date(
        d                               : TDateTime
    )                                   : string;

function str_date2hhmm(
        d                               : TDateTime
    )                                   : string;

function str_date2mdy(
        d                               : TDateTime
    )                                   : string;

function str_date2wwwdd(
        d                               : TDateTime
    )                                   : string;

function str_datetime(
        d                               : TDateTime
    )                                   : string;

function str_deb(
        s                               : string
    )                                   : string;

function str_delete(
        s                               : string;
        i                               : stt_int4;
        n                               : stt_int4
    )                                   : string;

procedure str_delimit(
        s                               : string;
        separator                       : string;
        delimiter                       : string;
    var n                               : stt_int4;
    var AoS                             : stt_AoS
    );

procedure str_delimit_TSL(
        s                               : string;
        separator                       : string;
        delimiter                       : string;
    var TSL                             : TStringList
    );

function str_downshift(
        s                               : string
    )                                   : string;

function str_fmtint(
        i                               : stt_int8;
        n                               : stt_int4
    )                                   : string;

function str_fmtlong(
        l                               : stt_int8;
        n                               : stt_int4
    )                                   : string;

function str_getbits(
        i                               : stt_card4;
        p                               : stt_int4;
        n                               : stt_int4
    )                                   : stt_card4;

function str_inet_ntoa(
        ip                              : stt_card4
    )                                   : string;

function str_int2commas(
        i                               : stt_int8
    )                                   : string;

function str_int2hex(
        i                               : stt_int8;
        width                           : stt_int4
    )                                   : string;

function str_int2str(
         i                              : stt_int8
    )                                   : string;

function str_IsNumeric(
        s                               : string
    )                                   : boolean;

function str_hex2int(
        s                               : string
    )                                   : stt_int8;

function str_justl(
         s                              : string;
         n                              : stt_int4
    )                                   : string;

function str_justr(
        s                               : string;
        n                               : stt_int4
    )                                   : string;

function str_len(
        s                               : string
    )                                   : stt_int4;

function str_lz(
    s                                   : string
    )                                   : string;

function str_nextid (
    var s                               : string;
        p                               : stt_int4;
        word_start                      : str_charset;
        word_elem                       : str_charset;
    var id_type                         : stt_int4;
    var id                              : string
   )                                    : stt_int4;

function str_pac2str(
    pac                                 : array of char
    )                                   : string;

function str_padl(
        s                               : string;
        n                               : stt_int4
    )                                   : string;

function str_padr(
        s                               : string;
        n                               : stt_int4
    )                                   : string;

function str_pos(
        s                               : string;
        substr                          : string
    )                                   : stt_int4;

function str_properCase(
        s                               : string;
        delimiters                      : str_charset
    )                                   : string;

function str_putbits(
        i                               : stt_card4;
        p                               : stt_int4;
        n                               : stt_int4;
        v                               : stt_card4
    )                                   : stt_card4;

function str_quotedstr (
    var s                               : string;
        p                               : stt_int4;
    var qs                              : string;
    var err                             : stt_int4
   )                                    : stt_int4;

function str_range(
        s                               : string;
        start                           : stt_int4;
        stop                            : stt_int4 = -1
   )                                    : string;

function str_real2commas(
        r                               : double;
        n                               : stt_int4
    )                                   : string;

function str_repeat(
        c                               : char;
        n                               : stt_int4
    )                                   : string;

function str_replace(
        s                               : string;
        f                               : string;
        t                               : string
    )                                   : string;

function str_rot13(
    s                                   : string
    )                                   : string;

function str_rot47(
        s                               : string
    )                                   : string;

function str_scan(
        s                               : string;
        pos                             : stt_int4;
        in_flag                         : boolean;
        cset                            : str_charset
   )                                    : stt_int4;

function str_setlen(
    var s                               : string;
        len                             : stt_int4
    )                                   : stt_int4;

function str_str2int(
        s                               : string
    )                                   : stt_int8;

function str_str2real(
        s                               : string
    )                                   : double;

procedure str_str2pac(
        s                               : string;
    var c                               : {packed} array of char
    );

procedure str_str2pac0(
        s                               : string;
    var pac                             : array of char
    );

function str_striphtml(
         s                              : string
   )                                    : string;

function str_substr(
        s                               : string;
        start                           : stt_int4;
        len                             : stt_int4
    )                                   : string;

function str_trim(
        s                               : string
    )                                   : string;

function str_triml(
        s                               : string
    )                                   : string;

function str_trimr(
        s                               : string
    )                                   : string;

function str_truncl(
        s                               : string;
        n                               : stt_int4
    )                                   : string;

function str_truncr(
         s                              : string;
         n                              : stt_int4
    )                                   : string;

function str_upshift(
    s                                   : string
    )                                   : string;


function str_unstring(
    s                                   : string;
    var strArr                          : array of string;
    sep                                 : string = ',';
    delim                               : string = ''
    )                                   : stt_int32b;

function str_wrap(
    s                                   : string;
    p                                   : stt_int4;
    len                                 : stt_int4
    )                                   : stt_int4;

// **********************************************************************

//-----------------------------------------------------------------------
implementation
//-----------------------------------------------------------------------

uses
    {$ifdef linux}
    process,
    {$endif}
    dos,
    math,
    strutils;

const
   debug                               =  false;

//-----------------------------------------------------------------------
//
// bool := fil_delete(fname);
//
//-----------------------------------------------------------------------

function fil_delete(
        fname                           : string
    )                                   : boolean;

begin

fil_delete := deletefile(fname);

end; // fil_delete

//-----------------------------------------------------------------------
//
// bool := fil_direxists(dirname);
//
//-----------------------------------------------------------------------

function fil_direxists(
        dirname                         : string
    )                                   : boolean;

begin

fil_direxists := DirectoryExists(dirname);

end; // fil_direxists


//-----------------------------------------------------------------------
//
// bool := fil_exists(fname);
//
//-----------------------------------------------------------------------

function fil_exists(
        fname                           : string
    )                                   : boolean;

begin

fil_exists := FileExists(fname);

end; // fil_exists

//-----------------------------------------------------------------------
//
// s := fil_getprogdir;
//
//-----------------------------------------------------------------------

function fil_getprogdir                 : string;

var
   s                                    : string;

begin

s := paramstr(0);
s := fil_qfname(s);

while (s[length(s)] <> '/') and (length(s) <> 0) do
    delete(s,length(s),1);

fil_getprogdir := s;

end; // fil_getprogdir

//-----------------------------------------------------------------------
//
// s := fil_qfname(fname);
//
//-----------------------------------------------------------------------

function fil_qfname(
        s                               : string
    )                                   : string;

begin

fil_qfname := FExpand(s);

end; // fil_qfname

//-----------------------------------------------------------------------
//
// fil_redirect;
//
//-----------------------------------------------------------------------

//  TurboPascal Does not allow I/O redirection from the DOS line. Calling
//  this procedure allows redirection. That's not a problem in GPC, so
//  this call has no effect.

procedure fil_redirect;

begin

end; // fil_redirect

//-----------------------------------------------------------------------
//
// err := fil_rename(old,new);
//
//-----------------------------------------------------------------------

function fil_rename(
        old                             : string;
        new                             : string
    )                                   : boolean;

begin

fil_rename := renamefile(old, new);

end; // fil_rename

//-----------------------------------------------------------------------
//
// l := fil_size(fname, err);
//
//-----------------------------------------------------------------------

function fil_size(
        fname                           : string;
    var err                             : stt_int4
    )                                   : stt_int8;

var
    f                                   : file of byte;

begin

err := 0;                                                   // errors no longer returned
assign(f, fname);
reset(f);
fil_size := filesize(f);
close(f);

end; // fil_size

//-----------------------------------------------------------------------
//
// fil_splitname(path, dir, name, ext);
//
//-----------------------------------------------------------------------

procedure fil_splitname(
        path                            : string;
    var dir                             : string;
    var name                            : string;
    var ext                             : string
   );

begin

dir  := extractFileDir(path);
name := extractFileName(path);
ext  := extractFileExt(path);

end; // fil_splitname

//-----------------------------------------------------------------------
//
// s := fil_tempfile(path);
//
//-----------------------------------------------------------------------

function fil_tempfile(
        path                            : string
    )                                   : string;

begin

result := GetTempFileName(path, '');

{

// OLD CODE

var
    filename                            :  string;
    i                                   :  stt_int8;
    ipath                               :  string;          // can't modify <path>

begin

ipath := str_triml(str_trimr(path));
if str_len(ipath) > 0 then
    if path[str_len(ipath)] <> '/' then
        ipath := ipath + '/';

repeat
    randomize;
    i := random(1000000);
    filename := str_truncl(str_fmtlong(i, 0),6) + '.tmp';
    until not fil_exists(ipath+filename);

fil_tempfile := ipath + filename;
}

end; // fil_tempfile


//-----------------------------------------------------------------------
//
// b := mai_SendHTML(tolist,cclist,bcclist,from,subject,msg);
//
//-----------------------------------------------------------------------

//  This procedure sets up a message to be sent in HTML format. You
//  provide the tolist (comma delimited), the cc&bcc list also comma
//  delimited, the from email address, a subject line and the contents
//  in an array of strings.

//  if an error occurs, this procedure returns true;

function mai_SendHTML(
        tolist                          : string;
        cclist                          : string;
        bcclist                         : string;
        from                            : string;
        subject                         : string;
        msgLineCount                    : stt_int4;
        msg                             : stt_aos
        )                               : boolean;

const
    MaxToAddrCount                      = 20;

var
    BCCArray                            : stt_AoS;
    BCCArrayCount                       : stt_int4;
    CCArray                             : stt_AoS;
    CCArrayCount                        : stt_int4;
    f                                   : text;
    k                                   : stt_int4;
    TempFileName                        : string;
    ToArray                             : stt_AoS;
    ToArrayCount                        : stt_int4;

    procedure sendit(
            msgaddr                     : string
        );

    var
        procOutput                      : string;
        s                               : string;

    begin

    {$ifdef linux}
        s:='sendmail ' + msgaddr + ' <' + TempFileName;
        if not RunCommand('/bin/sh', ['-c', s], procOutput) then
            raise EMaiRuncommand.Create('createProc of sendmail failed.');
    {$else}
        // when ready to implement, take a look at mailsend which I used in
        // earthshaker
        s := '';
        procOutput := '';
        raise EMaiWindows.Create('SendHTML not implemented in Windows.');
    {$endif}

    end; // sendit

begin {mai_SendHTML}

mai_SendHTML := true;

setlength(bccArray, maxToAddrCount);
setlength(ccArray,  maxToAddrCount);
setlength(toArray,  maxToAddrCount);


TempFileName := fil_tempfile('');
assign(f, tempFileName);
rewrite(f);

if str_len(ToList) > 0 then begin
    str_delimit(ToList, ',', '', ToArrayCount, ToArray);
    writeln(f, 'To: ', ToList);
    end
else
    ToArrayCount := 0;

if str_len(CCList) > 0 then begin
    str_delimit(CCList, ',', '', CCArrayCount, CCArray);
    writeln(f, 'cc: ', CCList);
    end
else
    CCArrayCount := 0;

if str_len(BCCList) > 0 then begin
    str_delimit(BCCList, ',', '', BCCArrayCount, BCCArray);
    writeln(f, 'bcc: ', BCCList);
    end
else
    BCCArrayCount := 0;

if str_len(from) > 0 then
    writeln(f, 'From: ', from);

if str_len(subject) > 0 then
    writeln(f, 'Subject: ', subject);

writeln(f, 'MIME-Version: 1.0');
writeln(f, 'Content-type: text/html');
writeln(f, 'Content-Transfer-Encoding: 7bit');
writeln(f);
writeln(f, '<html>');
writeln(f, '<body text="#000000" link="#0000ff" vlink="#551a8b" alink="#ff0000" ');
writeln(f, '    bgcolor="#ffffff">');
writeln(f, '<font face="Courier New" size="2">');
writeln(f, '<pre>');

{ output the message text }
for k := 1 to msgLineCount do begin
    writeln(f,str_AoSget(msg[k-1]));
    end;

{ close the temp file }
close(f);

{ Send the Message                                                            }
for k := 1 to ToArrayCount do begin
    sendit(str_AoSget(ToArray[k-1]));
    end; {for}

for k := 1 to CCArrayCount do begin
    sendit(str_AoSget(CCArray[k-1]));
    end; {for}

for k := 1 to BCCArrayCount do begin
    sendit(str_AoSget(BCCArray[k-1]));
    end; {for}

{ Purge the temp file                                                         }
fil_delete(TempFileName);

mai_SendHTML := false;

end;  {mai_SendHTML}

//-------------------------------------------------------------------------
//
// b := isbetween(x, lo, hi);
//
//-------------------------------------------------------------------------

//  returns true if lo <= x <= hi

function isbetween(
        x                               : stt_int8;
        loBound                         : stt_int8;
        hiBound                         : stt_int8
        )                               : boolean;

begin

if x < loBound then
    result := false
else if x > hiBound then
    result := false
else
    result := true;

end; // isbetween

//-------------------------------------------------------------------------
//
// memDump(v, sizeof(v));
//
// Calls memdump_tsl, then outputs string to stdlist.
//
// See memdump_tsl for full explanation.
//
//-------------------------------------------------------------------------

procedure memDump(
        v                               : pointer;
        size                            : stt_int32b
        );

var
    i                                   : stt_int32b;
    tsl                                 : TStringList       = nil;

begin

memDump_tsl(v, size, tsl);

for i := 0 to tsl.Count-1 do
    writeln(tsl.Strings[i]);

end; // memDump

//-------------------------------------------------------------------------
// memDump_tsl(v, sizeof(v), tsl);
//
// formatted dump of a variable in format of good ol' FCOPY
//
// For most variables you call using the '@' address operator:
//
// memDump_tsl(@array, sizeof(array), tsl);
//
// The above works for short strings, BUT NOT ansistrings.
//
// An ansistring is already a pointer, so you just type cast it as a pointer:
//
// memDump_tsl(pointer(string), str_len(string), tsl);
//
// ALL output is to a stringlist, TSL. The caller can then use writeln
// or a windows based algorith to display the output in tsl.
//-------------------------------------------------------------------------

procedure memDump_tsl(
        v                               : pointer;
        size                            : stt_int32b;
    var tsl                             : TStringList
    );


var
    ascStr                              : string;
    hexStr                              : string;
    i                                   : stt_int32b;
    p                                   : ^byte;

begin

try

    if tsl = nil then
        tsl := TStringList.Create;
    tsl.Clear;

    p := v;

    for i := 0 to size-1 do begin
        if i mod 20 = 0 then begin
            if i > 0 then begin
                // flush buffers
                tsl.Add(str_padr(hexStr,56) + ascStr);
                end;
            // init buffers
            hexStr := str_int2hex(i,4) + ': ';
            ascStr := '';
            end;
        // push next byte
        hexstr := hexstr + str_int2hex(p^,2);
        if (i+1) mod 2 = 0 then
            hexstr := hexstr + ' ';
        if isbetween(p^,32,128) then
            ascStr := ascStr + chr(p^)
        else
            ascStr := ascStr + '.';
        inc(p);
        end; // for

    tsl.add(str_padr(hexStr,56) + ascStr);

except
    on e:EAccessViolation do begin
        stackDump(e);
        raise;
        end;
    end; // try

end; // memDump_tsl

//-------------------------------------------------------------------------
// noop;
//
// Once in a blue moon, I need a NO-OP statement that does nothing but
// gives me a place to put a breakpoint.
//
//-------------------------------------------------------------------------

procedure noop;

begin

end;

//-------------------------------------------------------------------------
//
// s := pab2s(pab);
//
// Converts PAB to s. All trailing spaces and NULLs are stripped in the
// output string.
//
//-------------------------------------------------------------------------

function pab2s(
        pab                             : array of byte
    )                                   : string;

var
    i                                   : stt_int32b;
    p                                   : stt_int32b;

begin

// find first non blank, non null starting from the end

p := length(pab) - 1;
while (p >= 0) and (pab[p] in [byte(ord(' ')), 0]) do begin
    p := p - 1;
    end;

if p = -1 then begin
    result := '';
    exit;
    end;

result := '';
for i := 0 to p do
    result := result + char(pab[i]);

end; // pab2s

//-------------------------------------------------------------------------
//
// s2pab(pab, s);
//
// Converts s to PAB. If s is smaller than PAB, PAB will have trailing
// spaces added.
//
//-------------------------------------------------------------------------

procedure s2pab(
    var pab                             : array of byte;
        s                               : string
    );

var
    i                                   : stt_int32b;

begin

// init pab with spaces
for i := 0 to length(pab)-1 do
    pab[i] := byte(ord(' '));

// transfer s to pab
for i := 1 to min(length(pab), str_len(s)) do begin
    pab[i-1] := byte(ord(s[i]));
    end;

end; // s2pab

//-------------------------------------------------------------------------
//
// s2pab0(pab, s);
//
// Converts s to  null terminated PAB. If s is smaller than PAB, PAB will
// have trailing nulls.
// If len(s) >= sizeof(pab), s is truncated so the last byte in pab is the
// null.
//
//-------------------------------------------------------------------------

procedure s2pab0(
    var pab                             : array of byte;
        s                               : string
    );

var
    i                                   : stt_int32b;

begin

// init pab with zeros
for i := 0 to length(pab)-1 do
    pab[i] := 0;

// transfer s to pab
for i := 1 to min(length(pab)-1, str_len(s)) do begin
    pab[i-1] := byte(ord(s[i]));
    end;

end; // s2pab0

//-------------------------------------------------------------------------
//
// stackDump;
//
// Calls stackdump_tsl, then outputs string to stdlist.
//
// See stackdump_tsl for full explanation.
//
//-------------------------------------------------------------------------

procedure stackDump;

var
    i                                   : stt_int32b;
    tsl                                 : TStringList       = nil;

begin

stackDump_tsl(tsl);

for i := 0 to tsl.Count-1 do
    writeln(tsl.Strings[i]);
end;

//-------------------------------------------------------------------------
//
// stackDump(e);
//
// Calls stackdump_tsl, then outputs string to stdlist.
//
// See stackdump_tsl for full explanation.
//
//-------------------------------------------------------------------------

procedure stackDump(
        e                               : Exception
    );

var
    i                                   : stt_int32b;
    tsl                                 : TStringList       = nil;

begin

stackDump_tsl(e, tsl);

for i := 0 to tsl.Count-1 do
    writeln(tsl.Strings[i]);

end; // stackDump

//-------------------------------------------------------------------------
//
// stackDump_tsl(tsl);
//
// perform stack marker trace (a.k.a. back trace). You must pass the
// exception in that caused the error.
//
// Output is to TStringList TSL. This can then be used with writeln or
// and windows based output scheme.
//-------------------------------------------------------------------------

procedure stackDump_tsl(
    var tsl                             : TStringList
    );

var
    bp                                  : Pointer;
    callerAddress                       : Pointer;
    callerFrame                         : Pointer;
    i                                   : Longint;
    prevbp                              : Pointer;

const
    maxDepth                            = 20;

begin

if tsl = nil then
    tsl := TStringList.Create;
tsl.Clear;

tsl.Add('');
tsl.Add(
    '-------------------- S t a c k   M a r k e r   T r a c e --------------------');
tsl.Add('');

bp := get_frame;
try
    prevbp := bp - 1;
    i := 0;
    while bp > prevbp do begin
        callerAddress := get_caller_addr(bp);
        callerFrame := get_caller_frame(bp);
        if (callerAddress = nil) then
            break;
        tsl.Add(BackTraceStrFunc(CallerAddress));
        Inc(i);
        if (i >= maxDepth) or (callerFrame = nil) then
            break;
        prevbp := bp;
        bp     := callerFrame;
        end;
except
    { prevent endless dump if an exception occured }
    end; // try

end; // stackdump_tsl

//-------------------------------------------------------------------------
//
// stackDump_tsl(e, tsl);
//
// perform stack marker trace (a.k.a. back trace). You must pass the
// exception in that caused the error.
//
// Output is to TStringList TSL. This can then be used with writeln or
// and windows based output scheme.
//-------------------------------------------------------------------------

procedure stackDump_tsl(
        e                               : Exception;
    var tsl                             : TStringList
    );

var
    i                                   : stt_int32b;
    frames                              : PPointer;

begin

if tsl = nil then
    tsl := TStringList.Create;
tsl.Clear;

tsl.Add('');
tsl.Add(
    '-------------------- S t a c k   M a r k e r   T r a c e --------------------');
tsl.Add('');
if e <> nil then begin
    tsl.Add('Class: ' + E.ClassName + '. Message: ' + E.Message);
    end;

tsl.Add(BackTraceStrFunc(ExceptAddr));

frames := ExceptFrames;
for i := 0 to ExceptFrameCount - 1 do
    tsl.Add(BackTraceStrFunc(Frames[I]));

tsl.Add('');

end; // stackDump

 //-------------------------------------------------------------------------
 //
 // str_AoSdispose(AoS);
 //
 //-------------------------------------------------------------------------

 //  Given an array of string pointers, AoS, this procedure disposes of
 //  all of the strings.


procedure str_AoSdispose(
    var AoS                             : stt_AoS
    );

var
    k                                   : stt_int4;

begin //str_AoSdispose

for k := low(AoS) to high(AoS) do
   if AoS[k] <> nil then begin
       dispose(AoS[k]);
       Aos[k] := nil;                                       // ya'd think this was done automatically
       end;

end;  //str_AoSdispose

//-----------------------------------------------------------------------
//
// str_AoSdump(AoS);
//
//-----------------------------------------------------------------------

procedure str_AoSdump(
    var AoS                             : stt_AoS
   );


var
    k                                   : stt_int4;

begin //str_AoSdump

for k := low(AoS) to high(AoS) do begin
    write('  AoS[',k,'] = ');
    if aos[k] = nil then
        writeln('nil')
    else
        writeln(AoS[k]^);
    end;

end;  //str_AoSdump

//-----------------------------------------------------------------------
//
// str_AoSinit(AoS);
//
//-----------------------------------------------------------------------

//  This procedure will set all pointers in the Array of Strings <AoS>
//  to NIL.

// version 5.1 requires one use setlength to set the length of the aos a
// after declaration.

// instead of
//      aos : stt_aos(10)
// do
//      aos : stt_aos;
//      setlength(aos,10);

procedure str_AoSinit(
    var AoS                             : stt_aos
    );

var
    k                                   : stt_int4;

begin //str_AoSinit

for k := low(AoS) to high(AoS) do
    AoS[k] := nil;

end;  //str_AoSinit

//-----------------------------------------------------------------------
//
// s:= str_AoSget(AoS[n]);
//
//-----------------------------------------------------------------------

function str_AoSget(
    var AoSn                            : stt_pstr
    )                                   : string;

begin //str_AoSget

if AoSn = nil then
    str_AoSget := ''
else
    str_AoSget := AoSn^;

end;  //str_AoSget

//-----------------------------------------------------------------------
//
// str_AoSput(AoS[n],s);
//
//-----------------------------------------------------------------------

//  The following is not true in free pascal; however, using aosPut will
//  correctly call new on the intial string creation and it still really
//  the best option.

//  This procedure puts a string into an AoS element. You should never
//  directly assign strings such as this:
//
//    AoS[n]^ := <string var or string constant>
//
//  The reason is because when an AoS element is created, it is created
//  for the exact length of a string. If you try to put another string into
//  an existing AoS element, the new string will be truncated if it is longer
//  than the old string.
//
//  Use str_AoSput instead which will release the old string, create a new
//  element long enough for the new string, and then store it:
//
//     str_AoSput(AoS[n], 'This is a test');


procedure str_AoSput(
    var AoSn                            : stt_pstr;
        s                               : string
    );

begin //str_AoSput

if AoSn <> nil then begin
    dispose(AoSn);
    end;

new(AoSn);

if AoSn = nil then begin
    writeln('str_AoSput: new returned nil. s=',s);
    halt(1);
    end
else
    AoSn^ := s;

end;  //str_AoSput

//-----------------------------------------------------------------------
//
// b := str_contains(s1, s2);
//
// Returns true is s2 is found in s1. The search is case insensitive so
// str_contains('The quick brown fox','BROWN') will return true.
//-----------------------------------------------------------------------

function str_contains(
        s1                              : string;
        s2                              : string
    )                                   : boolean;

begin

if str_pos(str_upshift(s1), str_upshift(s2)) > 0 then
    result := true
else
    result := false;

end; // str_contains

//-----------------------------------------------------------------------
//
// n := str_count(s, substr);
//
//-----------------------------------------------------------------------

function str_count(
        s                               : string;
        substr                          : string
    )                                   : stt_int4;

var
    p                                   : stt_int4;

begin //str_count

if (str_len(s) < 1) or (str_len(substr) < 1) then begin
    result := 0;
    exit;
    end;

p := str_pos(s, substr);

if p = 0 then
    result := 0
else
    result := 1 + str_count(
        str_range(s, p+str_len(substr), str_len(s)),
        substr);

end;  //str_count

//-----------------------------------------------------------------------
//
//  s := str_date(now);
//  returns mm/dd/yyyy hh:mm:ss.mmmm
//
//-----------------------------------------------------------------------
function str_date(
        d                               : TDateTime
    )                                   : string;

var
    s                                   : string;
begin

s := FormatDateTime('z',d);
s := str_justr(s, 3);
s := str_lz(s);

result := FormatDateTime('mm/dd/yyyy hh:nn:ss', d) + '.' + s;

end; // str_date

//-----------------------------------------------------------------------
//
//  s := str_date2hhmm(now);
//  returns hh:mm:ss
//
//-----------------------------------------------------------------------
function str_date2hhmm(
        d                               : TDateTime
    )                                   : string;

begin

result := FormatDateTime('hh:nn:ss', d);

end; // str_date2hhmm

//-----------------------------------------------------------------------
//
//  s := str_date2mdy(now);
//  returns date as mm/dd/yyyy
//
//-----------------------------------------------------------------------
function str_date2mdy(
        d                               : TDateTime
    )                                   : string;

begin

result := FormatDateTime('mm/dd/yyyy', d);

end; // str_date2mdy

//-----------------------------------------------------------------------
//
//  s := str_date2wwwdd(now);
//  returns www dd such as MON 01
//
//-----------------------------------------------------------------------
function str_date2wwwdd(
        d                               : TDateTime
    )                                   : string;

begin

result := FormatDateTime('ddd dd', d);

end; // str_date2wwwdd

//-----------------------------------------------------------------------
//
//  s := str_datetime(now);
//  returns mm/dd/yyyy hh:mm:ss
//
//-----------------------------------------------------------------------
function str_datetime(
        d                               : TDateTime
    )                                   : string;

begin

result := FormatDateTime('mm/dd/yyyy hh:nn:ss', d);

end; // str_datetime

//-----------------------------------------------------------------------
//
// s := str_deb(s);
//
//-----------------------------------------------------------------------

function str_deb(
        s                               : string
    )                                   : string;

begin //str_deb

str_deb := str_trimr(str_triml(s));

end;  //str_deb

//-----------------------------------------------------------------------
//
// s := str_delete(s,i,n)
//
//-----------------------------------------------------------------------

function str_delete(
        s                               : string;
        i                               : stt_int4;
        n                               : stt_int4
    )                                   : string;

var
    temp                                : string;

begin //str_delete

if n = 0 then begin         //if # chars = 0, then don't delete anything
    str_delete := s;
    exit;
    end;

if i+n-1 > str_len(s) then   //if user tries to delete too much, reduce it
    n := str_len(s) - i + 1;
if (n < 0) or (i < 1) or (i > str_len(s)) then begin
    writeln('str_delete: bounds problem.');
    writeln('parm 1   : ', s);
    writeln('parm 2   : ', i);
    writeln('parm 3   : ', n);
    halt(1);
    end;

if str_len(s) = 1 then begin                                //handle simple 1 character string
    str_delete := '';
    exit;
    end;

temp := str_substr(s,1,i-1) + str_substr(s,i+n, str_len(s)-(i+n)+1);

str_delete := temp;

end;  //str_delete

//-----------------------------------------------------------------------
//
// str_delimit(s,separator,delimiter,n,AoS);
//
//-----------------------------------------------------------------------

//  Given <s>, <separator>, <delimiter>, this procedure breaks <s> into
//  multiple strings separated by <separator>. Each substring can be
//  separated by a delimited. If this delimiter is present, it will be
//  removed.

procedure str_delimit(
        s                               : string;
        separator                       : string;
        delimiter                       : string;
    var n                               : stt_int4;
    var AoS                             : stt_aos
);

var
    InString                            : boolean;
    p                                   : stt_int4;
    StartString                         : stt_int4;
    tmp                                 : string;

    procedure delimit(
        var  s                          : string;
            delimiter                   : string
        );

    begin //delimit

    if str_len(delimiter) = 0 then
        exit;

    if str_len(s) = 0 then
        exit;

    if s[1] = delimiter then
        s := str_delete(s,1,1);

    if s[str_len(s)] = delimiter then
        s := str_delete(s,str_len(s),1);

    end;  //delimit

begin //str_delimit

// initialize AoS
str_AoSinit(AoS);

// Look for parameter errors
if (str_len(separator) > 1) or (str_len(delimiter) > 1) then begin
    writeln('Error in str_delimit: separator and delimiter cannot be more than 1 character.');
    halt(1);
    end;

// If no string given, return nothing
if str_len(s) <= 0 then begin
    n := 0;
    exit;
    end;

// If no separator indicated, simply return the string
if str_len(separator) = 0 then begin
    n := 1;
    tmp  := str_deb(s);
    delimit(tmp, delimiter);
    str_AoSput(Aos[n-1],tmp);
    exit;
    end;

p             := 1;                 // tracks current position in <s>
StartString   := 1;                 // tracks last separator+1
n             := 0;
InString      := false;             // are we inside a delimited string?
while p <= str_len(s) do begin
    if InString then begin
        // inside a delimited string just watch for another delimiter
        if s[p] = delimiter then begin
            InString := false;
            end;
        end
    else if s[p] = delimiter then begin
        // delimiter found, now skip everything until another is found
        InString := true;
        end
    else if s[p] = separator then begin
        // separator found, store delimited string in AoS[n]
        n := n + 1;
        if n > high(aos)+1 then begin
            writeln('str_delimit: too many strings for return array');
            halt(1);
            end;
        tmp  := str_deb(str_range(s, StartString, p-1));
        delimit(tmp, delimiter);
        str_AoSput(Aos[n-1],tmp);
        StartString := p + 1;
        end;
    p := p + 1;
    end; //while

//store final string in AoS
n := n + 1;
if n > high(aos)+1 then begin
    writeln('str_delimit: too many strings for return array');
    halt(1);
    end;
tmp  := str_deb(str_range(s, StartString, p-1));
delimit(tmp, delimiter);
str_AoSput(Aos[n-1],tmp);

end;  //str_delimit

//-----------------------------------------------------------------------
//
// str_delimit_TSL(s,separator,delimiter,TStringList);
//
//-----------------------------------------------------------------------

//  like str_delimit except this puts delimited strings into a TStringList

procedure str_delimit_TSL(
        s                               : string;
        separator                       : string;
        delimiter                       : string;
    var TSL                             : TStringList
    );

var
    InString                            : boolean;
    p                                   : stt_int4;
    StartString                         : stt_int4;
    tmp                                 : string;

    procedure delimit(
        var  s                          : string;
            delimiter                   : string
        );

    begin //delimit

    if str_len(delimiter) = 0 then
        exit;

    if str_len(s) = 0 then
        exit;

    if s[1] = delimiter then
        s := str_delete(s,1,1);

    if s[str_len(s)] = delimiter then
        s := str_delete(s,str_len(s),1);

    end;  //delimit

begin  // str_delimit_TSL

if tsl = NIL then
    raise EStdUnit.Create('str_delimit_TSL: tsl is NIL.');

try
    tsl.Clear;
except
    writeln('str_delimit_TSL: unable to clear TStringList.');
    raise;
    end;

// Look for parameter errors
if (str_len(separator) > 1) or (str_len(delimiter) > 1) then begin
    writeln('str_delimite_TSL: separator and delimiter cannot be more than 1 character.');
    halt(1);
    end;

if str_len(s) <= 0 then                                     // If no string given, return nothing
    exit;

if str_len(separator) = 0 then begin                        // If no separator indicated, simply return the string
    tmp  := str_deb(s);
    delimit(tmp, delimiter);
    tsl.Add(tmp);
    exit;
    end;

p             := 1;                 // tracks current position in <s>
StartString   := 1;                 // tracks last separator+1
InString      := false;             // are we inside a delimited string?
while p <= str_len(s) do begin
    if InString then begin
        // inside a delimited string just watch for another delimiter
        if s[p] = delimiter then begin
            InString := false;
            end;
        end
    else if s[p] = delimiter then begin
        // delimiter found, now skip everything until another is found
        InString := true;
        end
    else if s[p] = separator then begin
        // separator found, store delimited string in AoS[n]
        tmp  := str_deb(str_range(s, StartString, p-1));
        delimit(tmp, delimiter);
        tsl.Add(tmp);
        StartString := p + 1;
        end;
    p := p + 1;
    end; //while

//store final string in AoS
tmp  := str_deb(str_range(s, StartString, p-1));
delimit(tmp, delimiter);
tsl.add(tmp);

end;  //str_delimit_TSL

//-----------------------------------------------------------------------
//
// s := str_downshift(s);
//
//-----------------------------------------------------------------------

function str_downshift(
        s                               : string
    )                                   : string;

var
    ch                                  : char;
    k                                   : stt_int4;

begin //str_downshift

for k := 1 to str_len(s) do begin
    ch := s[k];
    if (ord(ch)>=65) and (ord(ch)<=90) then
        s[k] := chr(ord(ch)+32);
    end;//for

str_downshift := s;

end;  //str_downshift

//-----------------------------------------------------------------------
//
// s := str_fmtint(i,n);
//
//-----------------------------------------------------------------------

function str_fmtint(
        i                               : stt_int8;
        n                               : stt_int4
    )                                   : string;

var
    s                                   : string;

begin //str_fmtint

if n = 0 then
    str(i, s)
else
    str(i:n, s);

str_fmtint := s;

end;  //str_fmtint

//-----------------------------------------------------------------------
//
// s := str_fmtlong(l, n);
//
//-----------------------------------------------------------------------

function str_fmtlong(
        l                               : stt_int8;
        n                               : stt_int4
    )                                   : string;

var
    s                                   : string;

begin //str_fmtlong

if n = 0 then
    str(l, s)
else
    str(l:n, s);

str_fmtlong := s;

end;  //str_fmtlong

//-----------------------------------------------------------------------------
//
// n := str_getbits(i, p, n);
//
// given a 32 bit <i>, a starting bit position <p>, and a number of bits
// <n>, this function extracts the bit field.
// if p or n are out of range, this routine returns $FFFFFFFF.
//-----------------------------------------------------------------------------

function str_getbits(
        i                               : stt_card4;
        p                               : stt_int4;
        n                               : stt_int4
    )                                   : stt_card4;

begin //str_getbits

if (p < 0) or (p > 31) or (n < 1) or (n > 32) or ((p + n) > 32) then begin
    result := $FFFFFFFF;
    exit;
    end;

// remove any left bits with a shift-left, shift right. "cardinal" overrides
// intermediate result.

i := cardinal(i shl p) shr p;

// shift all the way to the right to remove right bits

i := i shr (32-(p+n));

result := i;

end;  //str_getbits

//-----------------------------------------------------------------------------
//
// s := str_inet_ntoa(i);
//
// This function is like the mysql inet_ntoa: given a 32bit integer
// containing an IP address, it converts it to the normal 4 octet string form.
//-----------------------------------------------------------------------------

function str_inet_ntoa(
        ip                              : stt_card4
    )                                   : string;

var
    i                                   : stt_int4;

begin //str_inet_ntoa

result := '';
for i := 3 downto 0 do begin
    result := result + str_int2str((ip shr (i*8)) and $ff);
    if i <> 0 then
        result := result + '.';
    end; //for

end;  //str_inet_ntoa

//-----------------------------------------------------------------------
//
// s := str_int2commas(i);
//
//-----------------------------------------------------------------------

function str_int2commas(
        i                               : stt_int8
    )                                   : string;

var
    current                             : stt_int8;
    next3                               : string;
    remainder                           : stt_int8;
    s                                   : string;


begin //str_int2commas

{$warnings off}if debug then writeln('@str_int2commas(',i,')');{$warnings on}
if i = 0 then begin
    str_int2commas := '0';
    exit;
    end;

s := '';

current := abs(i);
while current > 0 do begin
    remainder := current mod 1000;
    current   := current div 1000;
    next3     := str_int2str(remainder);
    {$warnings off}if debug then writeln('@  remainder/current/next3 = ', remainder, '/', current, '/', next3);{$warnings off}
    if current > 0 then begin //more to go so pad with 0
        next3  := str_lz(str_padl(next3,3));
        s      := ',' + next3 + s;
        end
    else
        s      := next3 + s;
    end; //while

if i < 0 then
    s := '-' + s;

str_int2commas := s;

end;  //str_int2commas

//-----------------------------------------------------------------------------
//
// s := str_int2str(i);
//
//-----------------------------------------------------------------------------

function str_int2hex(
        i                               : stt_int8;
        width                           : stt_int4
    )                                   : string;

const
    hex                                 : array[0..$000F] of char = '0123456789ABCDEF';

begin //str_int2hex

if (width <= 1) and ((i and $FFFFFFF0) = 0) then
    str_int2hex := hex[i]
else
    str_int2hex := str_int2hex(cardinal(i) shr 4, width-1) + hex[i and $000F];

end;  //str_int2hex

//-----------------------------------------------------------------------
//
// s := str_int2str(i);
//
//-----------------------------------------------------------------------

function str_int2str(
        i                               : stt_int8
    )                                   : string;

begin //str_int2str

str(i, str_int2str);

end;  //str_int2str

//-----------------------------------------------------------------------
//
// b := str_isnumeric(s);
//
//-----------------------------------------------------------------------

function str_IsNumeric(
        s                               : string
    )                                   : boolean;

var
    k                                   : stt_int4;
    tmp                                 : string;

begin //str_IsNumeric

tmp := str_deb(s);

if str_len(tmp) < 1 then begin
    str_isnumeric := false;
    exit;
    end;

if tmp[1] = '-' then
    tmp := str_range(tmp, 2, str_len(tmp));

for k := 1 to str_len(tmp) do begin
    if not(tmp[k] in ['0'..'9']) then begin
        str_isnumeric := false;
        exit;
        end;
    end;

str_IsNumeric := true;

end;  //str_IsNumeric

//-----------------------------------------------------------------------------
//
// i := str_hex2int(s);
//
// returns hexadecimal string as decimal value. If a bad character is found
// This routing returns -1.
//-----------------------------------------------------------------------------

function str_hex2int(
        s                               : string
    )                                   : stt_int8;

const

hex                                     : array [0..15] of char = '0123456789ABCDEF';

var
k                                       : byte;
i                                       : stt_int8;

begin

str_hex2int := 0;

if str_len(s) = 0 then
    exit;

s := str_upshift(s);
for k := 1 to str_len(s) do begin
    i := 0;
    while (hex[i] <> s[k]) and (i < 16) do                  // compare letter
        i := i + 1;                                         // inc counter
    if i = 16 then begin
        str_hex2int := -1;                                  // -1 if invaid
        exit;
        end;
    str_hex2int := str_hex2int + i shl ((str_len(s) - k) * 4);// add to result
    end;

end;

//-----------------------------------------------------------------------
//
// s := str_justl(s,n);
//
//-----------------------------------------------------------------------

function str_justl(
        s                               : string;
        n                               : stt_int4
    )                                   : string;

var
    l                                   : stt_int4;

begin //str_justl

l := length(s);
if l > 0 then
    str_justl := str_padr(str_truncr(str_triml(str_trimr(s)),n),n)
else
    str_justl := str_repeat(' ',n);

end;  //str_justl

//-----------------------------------------------------------------------
//
// s := str_justr(s,n);
//
//-----------------------------------------------------------------------

function str_justr(
        s                               : string;
        n                               : stt_int4
    )                                   : string;

var
    l                                   : stt_int4;

begin //str_justr

l := length(s);
if l > 0 then
    str_justr := str_padl(str_truncl(str_triml(str_trimr(s)),n),n)
else
    str_justr := str_repeat(' ',n);

end;  //str_justr

//-----------------------------------------------------------------------
//
// s := str_len(s);
//
//-----------------------------------------------------------------------

function str_len(
        s                               : string
    )                                   : stt_int4;

begin //str_len

str_len := length(s);

end;  //str_len

//-----------------------------------------------------------------------
//
// s := str_lz(s);
//
//-----------------------------------------------------------------------

function str_lz(
        s                               : string
    )                                   : string;

var
    k                                   : stt_int4;

begin //str_lz

k := 1;
while (k < length(s)) and (s[k] = ' ') do begin
    s[k] := '0';
    k := k + 1;
    end;

str_lz := s;

end;  //str_lz

//-----------------------------------------------------------------------
//
// p := str_nextid(s,p,start,elem,type,id);
//
//-----------------------------------------------------------------------

//  This procedure returns the next id starting at s[p]. You pass a
//  string <s> containing multiple id's and <p>, an index into s.
//  Returned is <id_type>, describing the type of id found (see types
//  below with their corresponding id_type), and <id>, the id found
//  at s[p] and the function returns the index into <s> immediately
//  after the <id> (this can be used to find the next id.

//  An id can be a word(0), a number(1), a string(2), or a special
//  character(3). If the string contains no symbols, id_type is
//  returned as -1. If an unmatched string is found, (-2) is returned.

//  A word starts with any element of <word_start> and contains only
//  elements from <word_elem>.

//  A number starts with a digit or '%' and can contain only digits.
//  If <word_start> contains '%' then '%' cannot be used in numbers.

//  A string starts with ' or " and is terminated with the same
//  character. it can contain any character within it. If <word_start
//  contains ' or " then strings cannot exist.

//  a special character is any character not found in <word_start>,
//  not '%' and not ' or ".

//  handling blanks: a blank can seperate ids, but is not an id
//  itself (not even a special character).

//  Example: the string 'this is a test "of whatever" + 123' would return:
//      id_type     id
//      0        this
//      0        is
//      0        a
//      0        test
//      2        of whatever
//      3        +
//      1        123

function str_nextid (
    var s                               : string;
        p                               : stt_int4;
        word_start                      : str_charset;
        word_elem                       : str_charset;
    var id_type                         : stt_int4;
    var id                              : string
    )                                   : stt_int4;

label
    999;

var
    k                                   : stt_int4;
    loop                                : boolean;

begin //str_nextid

if p > str_len (s) then begin                               // handle <p> pointing past end of string
    id_type := -1;
    str_nextid := p;
    goto 999;
    end;

{$warnings off}if debug then writeln('@str_nextid: entry, s=', copy(s,p,str_len(s)-p+1));{$warnings on}

loop := true;
while loop do begin
    if s[p] = ' ' then begin
        p := p + 1;
        if p > str_len (s) then begin
            loop := false;
            id_type := -1;
            end;
        end
    else if (s[p] = '"') or (s[p] = '''') then begin
        id_type := 2;
        id      := '';
        k       := p + 1;
        while (k <= str_len (s)) and (s[k]<>s[p]) do begin
            id := id + copy(s, k, 1);
            k  := k + 1;
            if k > sizeof(id) then
                k := str_len(s) + 1;
            end;
        loop := false;
        if k <= str_len(s) then
            p := k + 1
        else begin
            p := k;
            id_type := -2;
            end;
        end
    else if s[p] in word_start then begin
        id_type := 0;
        id      := copy(s, p, 1);
        k       := p + 1;
        while (k <= str_len (s)) and (s[k] in word_elem) do begin
            id := id + copy(s, k, 1);
            k  := k + 1;
            if k > sizeof(id) then
                k := str_len(s) + 1;
            end;
        loop := false;
        p    := k;
        end
    else if s[p] in ['0'..'9','%'] then begin
        id_type := 1;
        id      := copy(s, p, 1);
        k       := p + 1;
        while (k <= str_len (s)) and (s[k] in ['0'..'9']) do begin
            id := id + copy(s, k, 1);
            k  := k  + 1;
            if k > sizeof(id) then
                k := str_len(s) + 1;
            end;
        loop := false;
        p    := k;
        end
    else begin
        id_type := 3;
        id      := copy(s, p, 1);
        loop    := false;
        p       := p + 1;
        end;
    end; //loop

str_nextid := p;

999: end;  //str_nextid

//-----------------------------------------------------------------------
//
// s := str_pac2str(pac);
//
//-----------------------------------------------------------------------

function str_pac2str(
    pac                                 : array of char
    )                                   : string;

var
    i                                   : stt_int32b;

begin

//writeln('@pac2str');

result := '';

i := 0;
while (i < length(pac)) and (ord(pac[i]) <> 0) do begin
    result := result + pac[i];
    i      := i + 1;
    end; // while

//writeln('  result: ', result);

end; // str_pac2str

//-----------------------------------------------------------------------
//
// s := str_padl(s,n);
//
//-----------------------------------------------------------------------

function str_padl(
        s                               : string;
        n                               : stt_int4
    )                                   : string;

begin //str_padl

if (n < 0) or (n > 255) then begin
    writeln('str_padl: n out of bounds. n = ', n);
    halt(1);
    end;

if length(s) < n then
    str_padl := concat(str_repeat(' ',n-length(s)),s)
else
    str_padl := s;

end;  //str_padl

//-----------------------------------------------------------------------
//
// s := str_padr(s,n);
//
//-----------------------------------------------------------------------

function str_padr(
        s                               : string;
        n                               : stt_int4
    )                                   : string;

begin //str_padr

if (n < 0) or (n > 255) then begin
    writeln('str_padr: n out of bounds. n = ', n);
    halt(1);
    end;

if length(s) < n then
    str_padr := concat(s, str_repeat(' ',n-length(s)))
else
    str_padr := s;

end;  //str_padr

//-----------------------------------------------------------------------
//
// p := str_pos(s,substr);
//
//-----------------------------------------------------------------------

//  This function is in every pascal but is always seems to be inconsistent
//  so I'm implementing my own method to call it. If <substr> is not found
//  in <s>, then 0 is returned.

function str_pos(
        s                               : string;
        substr                          : string
    )                                   : stt_int4;

var
    p                                   : stt_int4;

begin //str_pos

p       := pos(substr,s);

str_pos := p;

end;  //str_pos

//-----------------------------------------------------------------------
//
// s := str_properCase(s, delimiters);
//
//-----------------------------------------------------------------------

//  This function formats a string into 'proper' casing which means all
//  lowercase except first letter of each word.

//  delimiters is a list of delimiters used to determine where a new
//  word start.

//  Currently this is just a wrapper for free pascal AnsiProperCase, but
//  I want my own version just in case I port it to another pascal.

function str_properCase(
        s                               : string;
        delimiters                      : str_charset
    )                                   : string;

begin

result := AnsiProperCase(s, delimiters);

end; // str_properCase

//-----------------------------------------------------------------------------
//
// n := str_putbits(i, p, n, v);
//
// given a 32 bit <i>, a starting bit position <p>, a number of bits <n>, and
// a value <v>, this function stores <v> in the requested bit field.
// if p, n or v are out of range, this routine returns 16#FFFFFFFF.
//-----------------------------------------------------------------------------

function str_putbits(
        i                               : stt_card4;
        p                               : stt_int4;
        n                               : stt_int4;
        v                               : stt_card4
    )                                   : stt_card4;

var
mask                                    : cardinal;

begin //str_putbits

if (p < 0) or (p > 31) or (n < 1) or (n > 32) or ((p + n) > 32) then begin
    writeln('str_putbits: p or n value bad. p=', p, ' n=', n);
    result := $FFFFFFFF;
    exit;
    end;

if (v > intpower(2,n)-1) then begin
    writeln('str_putbits: value ', v, ' won''t fit in ', n, ' bits.');
    result := $FFFFFFFF;
    exit;
    end;

// Build a mask to clear only the affected bits
mask := not(cardinal(trunc(2**n)-1) shl (32-(p+n)));

// Clear all bits in the bit field to be updated
i := i and mask;

// Shift new value into the correct bit position
v := v shl (32-(p+n));

// Insert new value into bit field
result := i or v;

writeln('i = ', i, ' mask = ', mask, ' v = ', v);

end;  //str_putbits

//-----------------------------------------------------------------------
//
// p := str_quotedstr(s, p, qs, err);
//
//-----------------------------------------------------------------------

//  This procedure is used to extract a quoted string from <s>. It
//  assumes <p> points to the starting quote (or leading blanks).
//  once the leading quote (' or ") is found, the matching closing
//  quote is found. you are allowed to imbed the quote mark by
//  placing double quote marks in the string (eg 'a''bc' or '''').
//  the string (without the quotes) is placed in <qs>. if an error
//  occurs, it is returned in <err>. <err> = 1 means there is no starting
//  quote while = 2 means there is a quote nesting problem.
//  <p> is returned as the next character after the last quote mark.
//  you may also nest quotes such as "'"..."'" in which case you are
//  returned '"..."'. also, imbedded quotes are ignored inside of
//  nested quotes. "joe""s sign says 'i can''t work'" is returned as
//  joe"s sign says 'i can''t work'.

function str_quotedstr (
    var s                               : string;
        p                               : stt_int4;
    var qs                              : string;
    var err                             : stt_int4
)                                       : stt_int4;

label
    999;

const
    er_expected_quote                   = 1;
    er_quote_nest                       = 2;

var
    ch                                  : string[1];
    loop                                : boolean;
    q                                   : packed array [1 .. 32] of char;
    tos                                 : stt_int4;

begin //str_quotedstr

err := 0;

p := str_scan (s, p, true, [' ']);
if (p > str_len(s)) or not((s[p] = '''') or (s[p] = '"')) then begin
    err := er_expected_quote;
    goto 999;
    end;

tos  := 0;
qs   := '';
loop := true;
while loop do begin

    if p > str_len(s) then begin
        err  := er_quote_nest;
        goto 999;
        end;

    ch := copy(s, p, 1);
    p  := p + 1;

    if (ch = '''') or (ch = '"') then begin
        qs := qs + ch;
        if tos = 0 then begin
            tos := tos + 1;
            q[tos] := ch[1];
            end
        else begin
            if ch[1] = q[tos] then begin
                // if double quotes exist, then you merely copy the next
                // quote. this means "hi ""you"" happening dude" becomes
                // <hi "you" happening dude>.
                if (tos = 1) and (p <= str_len(s)) and (s[p] = ch[1]) then
                    p := p + 1
                else begin
                    tos := tos - 1;
                    if tos = 0 then
                        loop := false;
                    end;
                end
            else begin
                tos := tos + 1;
                if tos > sizeof(q) then begin
                    err := er_quote_nest;
                    goto 999;
                    end;
                q[tos] := ch[1];
                end;
            end;
        end
    else
        qs := qs + ch;

    end; //while

qs := copy(qs,2,str_len(qs)-2);

999:
if err > 0 then
    qs := '';
str_quotedstr := p;

end;  //str_quotedstr

//-----------------------------------------------------------------------
//
// s := str_range(s, start, stop);
//
// As of version 5.7, stop is now optional in which case -1 is the default.
// When stop is -1, it is the same as str_len(s), so writing
// str_range(s, 1) would be the same as str_range(s, 1, str_len(s)).
//
//-----------------------------------------------------------------------

function str_range(
        s                               : string;
        start                           : stt_int4;
        stop                            : stt_int4      = -1
    )                                   : string;

begin //str_range

if (stop = -1) then
    stop := str_len(s);

if (start < 1) or (stop < 1) or (start > str_len(s)) or (start > stop) then begin
    str_range := '';
    end;

if stop > str_len(s) then
    stop := str_len(s);

str_range := copy(s, start, stop-start+1);

end;  //str_range

//-----------------------------------------------------------------------
//
// s := str_real2commas(r,n);
//
//-----------------------------------------------------------------------

//  Given a real value (long real), and the number of digits to maintain
//  after the decimal point (n), this procedure converts r to a string
//  in the form nnn,nnn.nn.

function str_real2commas(
        r                               : double;
        n                               : stt_int4
    )                                   : string;

var
    fracpart                            : double;
    k                                   : stt_int4;
    s                                   : string;
    wholepart                           : stt_int8;

begin //str_real2commas

{$warnings off}if debug then writeln('@str_real2commas(',r,',',n,')');{$warnings on}

if n < 0 then
    n := 0;

r := r + ((5 / intpower(10,(n+1)))); // round to the correct position based on <n>
{$warnings off}if debug then
    writeln('@  after rounding (added ', 5 / intpower(10,(n+1)),') r=', r);{$warnings on}

wholepart := trunc(r);
fracpart  := abs(frac(r));
{$warnings off}if debug then writeln('@  whole/frac=', wholepart, '/', fracpart);{$warnings on}

// formatting the whole part is a piece of cake
s := str_int2commas(wholepart) + '.';

// shift each digit before the decimal point, and append it to the string
for k := 1 to n do begin
    fracpart := fracpart * 10;
    s := s + str_int2str(trunc(fracpart));
    fracpart := frac(fracpart);
    end; //for

str_real2commas := s;

end;  //str_real2commas

//-----------------------------------------------------------------------
//
// s := str_repeat(c,n);
//
//-----------------------------------------------------------------------

function str_repeat(
        c                               : char;
        n                               : stt_int4
    )                                   : string;

var
    k                                   : stt_int4;
    s                                   : string;
    s2                                  : string[1];

begin //str_repeat

if (n < 1) or (n > 255) then begin
    writeln('str_repeat: n out of bounds. n = ', n);
    halt(1);
    end;

s  := '';
s2 := c;
for k := 1 to n do
    s := concat(s, s2);

str_repeat := s;

end;  //str_repeat

//-----------------------------------------------------------------------
//
// s := str_replace(s,f,t);
//
//-----------------------------------------------------------------------

//  Replaces all occurrences of <f> with <t> in <s>, returning the
//  resulting string.

//  if <t> is null, then the <f> string is deleted.

//    str_replace('12345', '3', '') => '1245'

//  If <t> is larger than <f>, then the effect is that <f> is deleted
//  and <t> inserted:

//  str_replace('12345', '3', 'xxx') => '12xxx45'

//  When <f> is found, it is replaced with <t> and replacement starts again
//  on the next character after the location where <f> was found:

//  str_replace('1234512345', '3', 'x') => '12x4512x45'

//  This allows <t> to contain <f>:

//  str_replace('1234512345', '3', '83') => '128345128345'

function str_replace(
        s                               : string;
        f                               : string;
        t                               : string
    )                                   : string;

var
    p                                   : stt_int4;

begin //str_replace

{$warnings off}if debug then writeln('@str_replace, s/f/t=', s, '/', f, '/', t);{$warnings on}

if str_len(s) < 1 then begin
    str_replace := '';
    exit;
    end;

if str_len(f) < 1 then begin
    str_replace := s;
    exit;
    end;

p := str_pos(s, f);

{$warnings off}if debug then writeln('@str_replace, p=', p);{$warnings on}

if p = 0 then begin
    //<f> not found, simply return <s>
    {$warnings off}if debug then writeln('  p=0 executed');{$warnings on}
    result := s;
    end
else if p = 1 then begin
    //<f> found at start of <s>
    {$warnings off}if debug then writeln('  p=1 executed');{$warnings on}
    result := t + str_replace(str_range(s, str_len(f)+1, str_len(s)),f,t);
    end
else begin//<f> found in middle/end of <s>
    {$warnings off}if debug then writeln('  p = middle executed');{$warnings on}
    result := str_range(s, 1, p-1) + t +
    str_replace(str_range(s,p+str_len(f),str_len(s)),f,t);
    end;

{$warnings off}if debug then writeln('@str_replace, on exit, result=', result);{$warnings on}

end;  //str_replace

//-----------------------------------------------------------------------
//
// s := str_rot13(s);
//
//-----------------------------------------------------------------------

//  returns rot13 encoding of string

function str_rot13(
        s                       : string
        )                       : string;

var
    x                           : integer;
    c                           : char;
    b                           : byte;
//      s                           : string;

begin

//Store data.
//s := Source;

for x := 1 to Length(s) do begin
    //Get char
    c := s[x];
    //Get byte
    b := Ord(c);

    if (c >= 'A') and (c <= 'M') or (c >= 'a') and (c <= 'm') then begin
        //Plus 13
        Inc(b, 13);
        end;

    if (c >= 'N') and (c <= 'Z') or (c >= 'n') and (c <= 'z') then begin
        //Minus 13
        Dec(b, 13);
        end;

    //Build string
    s[x] := chr(b);
    end; // for

Result := s;

end; // str_rot13

//-----------------------------------------------------------------------
//
// s := str_rot47(s);
//
//-----------------------------------------------------------------------

//  like str_rot13, except encodes digits and special chars

function str_rot47(
        s                       : string
        )                       : string;

var
    i                           : integer;
    j                           : integer;

begin

Result := s;
for i := 1 to Length(s) do begin
    j := Ord(s[i]);
    if (j in [33..126]) then begin
        Result[i] := Chr(33 + ((j + 14) mod 94));
        end;
    end; // for i

end; // str_rot47

//-----------------------------------------------------------------------
//
// i := str_scan(s,i,inflag,cset);
//
//-----------------------------------------------------------------------

//  This procedure scans the string <s> starting at <pos>. It stops
//  scanning if <in_flag> is true and s[pos] not in <cset> or if
//  in_flag is false and s[pos] in <cset>. str_scan returns the position
//  in <s> where s[pos] no longer matches <cset> (or begins matching
//  <cset>). If the end of string is hit before the terminator is found,
//  str_scan returns str_len(s) + 1.

//  Examples:
//  given s := 'this is a test 1234';
//  str_scan(s, 1, true, ['a'..'z']) returns 5 (finds "this")
//  str_scan(s, 1, false, ['0'..'9']) returns 16 (finds "1").

//  You can use str_scan with copy to extract "tokens". For example,
//  assuming <p> points to the beginning of a token, use:
//  token := copy(s, p, str_scan(s, p, true, ['A'..'Z','a'..'z','0'..'9'])-p);

function str_scan(
        s                               : string;
        pos                             : stt_int4;
        in_flag                         : boolean;
        cset                            : str_charset
    )                                   : stt_int4;

var
    k                                   : stt_int4;

begin //str_scan

k := pos;
if in_flag then
    while (k <= str_len(s)) and (s[k] in cset) do
        k := k + 1
else
    while (k <= str_len(s)) and not(s[k] in cset) do
        k := k + 1;

str_scan := k;

end;  //str_scan

//-----------------------------------------------------------------------
//
// i := str_setlen(s,len);
//
//-----------------------------------------------------------------------

function str_setlen(
    var s                               : string;
        len                             : stt_int4
    )                                   : stt_int4;

begin //str_setlen

if (len < 0) then begin
    writeln;
    writeln('str_setlen: requested length out of range. Must be >= 0 ');
    writeln('program requested ', len);
    writeln;
    halt(1);
    end;

str_setlen := length(s);
setlength(s, len);

end;  //str_setlen

//-----------------------------------------------------------------------
//
// i := str_str2int(s);
//
//-----------------------------------------------------------------------

//  converts string to integer. If <s> is not an integer then this proc
//  simply returns 0 SO TEST IT IF YOU AREN'T SURE.

function str_str2int(
        s                               : string
    )                                   : stt_int8;

begin //str_str2int

val(s, str_str2int);

end;  //str_str2int

//-----------------------------------------------------------------------
//
// r := str_str2int(s);
//
//-----------------------------------------------------------------------

//  converts string to real. If <s> is not an integer then this proc
//  simply returns 0 SO TEST IT IF YOU AREN'T SURE.

function str_str2real(
        s                               : string
    )                                   : double;

begin //str_str2real

val(s, str_str2real);

end;  //str_str2real


//-----------------------------------------------------------------------
//
// str_str2pac(s, pac);
//
//-----------------------------------------------------------------------

//  string to packed array of char assignment doesn't pad with trailing
//  spaces (bizzare if you ask me). So this procedure will assign a string
//  to a packed array of char and fill it with trailing spaces.

//  NOTE: it has perplexed my why this procedure works - how does LENGTH
//  know the length of (c)???

//  Today I see that packed array of char is simply a C type string - null
//  terminated. So if you have packed array [1 ..30] of char and assign
//  any string w/ a length < 30, str_str2pac will work. HOWEVER, I believe
//  that if a string of length 30 is assigned, there is no terminating null
//  and this procedure will fail.

//  NOPE! I don't know how it is doing it, but it does have the proper
//  length.

procedure str_str2pac(
        s                               : string;
    var c                               : packed array of char
    );

var
    i                                   : integer;

begin

if str_len(s) >= length(c) then
    //if len of s is >= c, there will be no padding, so just copy
    //(and possibly truncate).
    for  i:=1 to length(c) do c[i-1]:=s[i]
else begin
    //copy s to c, then pad with trailing spaces
    for i:=1 to str_len(s) do c[i-1]:=s[i];
    for i:=str_len(s)+1 to length(c) do c[i-1]:=' ';
    end;

end;

//-----------------------------------------------------------------------
//
// str_str2pac0(s, pac);
//
//-----------------------------------------------------------------------

//  Like str_str2pac0, except ends with trailing #0 if PAC isn't filled.

procedure str_str2pac0(
        s                               : string;
    var pac                             : array of char
    );

var
    i                                   : stt_int32b;

begin

if str_len(s) >= length(pac) then
    //if len of s is >= c, there will be no padding, so just copy
    //(and possibly truncate).
    for  i := 1 to length(pac) do
        pac[i-1] := s[i]
else begin
    //copy s to c, then pad with trailing nulls
    for i := 1 to str_len(s) do
        pac[i-1] := s[i];
    for i := str_len(s)+1 to length(pac) do
        pac[i-1] := #0;
    end;

end;

//-----------------------------------------------------------------------
//
// s := str_striphtml(s);
//
//-----------------------------------------------------------------------

//  strips all HTML out of the string. Essentially, if '<' is found,
//  everything to the next '>' is ignored.

function str_striphtml(
        s                               : string
    )                                   : string;

var
    p                                   : stt_int4;
    q                                   : stt_int4;


begin //str_striphtml

result := s;

while true do begin
    p := str_pos(result,'<');
    if p = 0 then
        break;
    q := str_pos(str_range(result,p,str_len(result)),'>');
    if q = 0 then
        break;
    result := str_delete(result,p,q);
    end;

end;  //str_striphtml

//-----------------------------------------------------------------------
//
// s := str_substr(s, start, len);
//
//-----------------------------------------------------------------------

function str_substr(
        s                               : string;
        start                           : stt_int4;
        len                             : stt_int4
    )                                   : string;

begin //str_substr

str_substr := copy(s, start, len);

end;  //str_substr

//-----------------------------------------------------------------------
//
// s := str_trim(s);
//
//-----------------------------------------------------------------------

function str_trim(
        s                               : string
    )                                   : string;

begin

str_trim := str_deb(s);

end;

//-----------------------------------------------------------------------
//
// s := str_triml(s);
//
//-----------------------------------------------------------------------

function str_triml(
        s                               : string
    )                                   : string;

var
    k                                   : stt_int4;

begin //str_triml

str_triml := '';
if length(s) < 1 then
    exit;
k := 1;
while (k < length(s)) and (s[k] = ' ') do
    k := k + 1;
str_triml := copy(s, k, length(s)-k+1);

end;  //str_triml

//-----------------------------------------------------------------------
//
// s := str_trimr(s);
//
//-----------------------------------------------------------------------

function str_trimr(
        s                               : string
    )                                   : string;

var
    k                                   : stt_int4;

begin //str_trimr

str_trimr := '';
if length(s) < 1 then
    exit;
k := length(s);
while (k > 0) and (s[k] in [' ',#0]) do
    k := k - 1;
delete(s,k+1,length(s)-k);
str_trimr := s;

end;  //str_trimr

//-----------------------------------------------------------------------
//
// s := str_truncl(s,n);
//
//-----------------------------------------------------------------------

function str_truncl(
        s                               : string;
        n                               : stt_int4
    )                                   : string;

begin //str_truncl

if (n < 0) or (n > 255) then begin
    writeln('str_truncl: n out of bounds. n = ', n);
    halt(1);
    end;

if length(s) > n then
    str_truncl := copy(s, length(s)-n+1, n)
else
    str_truncl := s;

end;  //str_truncl

//-----------------------------------------------------------------------
//
// s := str_truncr(s,n);
//
//-----------------------------------------------------------------------

function str_truncr(
        s                               : string;
        n                               : stt_int4
    )                                   : string;

begin //str_truncr

if (n < 0) or (n > 255) then begin
    writeln('str_truncr: n out of bounds. n = ', n);
    halt(1);
    end;

if length(s) > n then
    str_truncr := copy(s, 1, n)
else
    str_truncr := s;

end;  //str_truncr

//-----------------------------------------------------------------------
//
// s := str_upshift(s);
//
//-----------------------------------------------------------------------

function str_upshift(
        s                               : string
    )                                   : string;

var
    ch                                  : char;
    k                                   : stt_int4;

begin //str_upshift

for k := 1 to str_len(s) do begin
    ch := s[k];
    if (ord(ch)>=97) and (ord(ch)<=122) then
        s[k] := chr(ord(ch)-32);
    end; //while

str_upshift := s;

end;  //str_upshift

//-----------------------------------------------------------------------
//
// n := str_unstring(s, strArr[, sep[, delim]]);
//
//-----------------------------------------------------------------------

function str_unstring(
    s                                   : string;
    var strArr                          : array of string;
    sep                                 : string = ',';
    delim                               : string = ''
    )                                   : stt_int32b;

const
    maxstr                              = 100;

var
    aos                                 : stt_aos;
    k                                   : stt_int32b;

begin

result := 0;
setlength(aos,maxstr);
str_delimit(s, sep, delim, result, aos);

if result > length(strArr) then
    result := length(strArr);

for k := 0 to result-1 do
    strArr[k] := str_AoSget(aos[k]);

str_aosdispose(aos);

end; // str_unstring

//-----------------------------------------------------------------------
//
// i := str_wrap(s, p, len);
//
//-----------------------------------------------------------------------
{
    Given string s and starting location p, we want to find the location
    to split the string nearest to s+p+len, by going backwards looking for
    a space. <i> will point to the first space at or before p+len.

    There may well be spaces before and after s[i], so you still need to
    account for them.

    If str_len(s) - (p-1) < len (e.g. the rest of the string fits), then
    str_len(s)+1 is returned.

    If no space is found before parsing all the way back to s[p], then
    0 is returned.

    bounds issues will return a -1;

    if a single word (e.g. anything delimited by spaces) is > len, then p+len
    is returned which doesn't point to a space, it points to the last
    character in s that will fit onto the line, causing the word to be split.
    The caller could detect this by seeing that s[x] <> ' ' where x is the
    value returned.
}

function str_wrap(
    s                                   : string;
    p                                   : stt_int4;
    len                                 : stt_int4
    )                                   : stt_int4;

var
    i                                   : stt_int4;

begin

if (p < 1) then begin
    result := -1;
    exit;
    end;
if (len < 1) then begin
    result := -1;
    exit;
    end;
if (str_len(s) < 1) then begin
    result := 1;                                            // if nullstr, point to 0+1
    exit;
    end;
if (str_len(s) - p < len) then begin
    // s[p:] fits within len
    result := str_len(s) + 1;
    exit;
    end;

i := p + len; // i points to 1 past max output size
while true do begin
    if i = p then begin // no spaces found
        result := p + len;
        exit;
        end;
    if s[i] = ' ' then begin
        result := i;
        exit;
        end;
    i := i - 1;
    end; // while

end; // str_wrap


end.


